from typing import cast

import requirements
from crunch_convert._model import RequirementLanguage
from crunch_convert.requirements_txt._model import NamedRequirement


def parse_from_file(
    *,
    language: RequirementLanguage = RequirementLanguage.PYTHON,
    file_content: str
):
    return [
        NamedRequirement(
            name=cast(str, requirement.name),
            extras=requirement.extras,
            specs=[
                "".join(spec_parts)
                for spec_parts in requirement.specs
            ],
            language=language,
        )
        for requirement in requirements.parse(file_content)
    ]



__title__ = 'crunch-convert'
__description__ = 'crunch-convert - Conversion module for the CrunchDAO Platform'
__version__ = '0.4.0'
__author__ = 'Enzo CACERES'
__author_email__ = 'enzo.caceres@crunchdao.com'
__url__ = 'https://github.com/crunchdao/crunch-convert'



class MockedWriteTextIO:

    def write(self, *args, **kwargs):  # type: ignore
        pass  # pragma: no cover

    def __enter__(self):  # type: ignore
        return self  # pragma: no cover

    def __exit__(self, exc_type, exc_value, traceback):  # type: ignore
        pass  # pragma: no cover



from crunch_convert._model import RequirementLanguage as RequirementLanguage
from crunch_convert.requirements_txt._formatter import format_files_from_imported as format_files_from_imported
from crunch_convert.requirements_txt._formatter import format_files_from_named as format_files_from_named
from crunch_convert.requirements_txt._freeze import CrunchHubVersionFinder as CrunchHubVersionFinder
from crunch_convert.requirements_txt._freeze import LocalSitePackageVersionFinder as LocalSitePackageVersionFinder
from crunch_convert.requirements_txt._freeze import VersionFinder as VersionFinder
from crunch_convert.requirements_txt._freeze import freeze as freeze
from crunch_convert.requirements_txt._model import NamedRequirement as NamedRequirement
from crunch_convert.requirements_txt._parse import parse_from_file as parse_from_file
from crunch_convert.requirements_txt._whitelist import CachedWhitelist as CachedWhitelist
from crunch_convert.requirements_txt._whitelist import CrunchHubWhitelist as CrunchHubWhitelist
from crunch_convert.requirements_txt._whitelist import Library as Library
from crunch_convert.requirements_txt._whitelist import LocalWhitelist as LocalWhitelist
from crunch_convert.requirements_txt._whitelist import MultipleLibraryAliasCandidateException as MultipleLibraryAliasCandidateException
from crunch_convert.requirements_txt._whitelist import Whitelist as Whitelist



from abc import ABC, abstractmethod
from typing import List, Literal, Optional, overload

import requests

from crunch_convert._model import RequirementLanguage
from crunch_convert.requirements_txt._model import NamedRequirement
from crunch_convert.requirements_txt._whitelist import Whitelist


@overload
def freeze(
    requirements: List[NamedRequirement],
    *,
    freeze_only_if_required: Literal[True],
    whitelist: Whitelist,
    version_finder: "VersionFinder",
) -> List[NamedRequirement]:
    ...  # pragma: no cover


@overload
def freeze(
    requirements: List[NamedRequirement],
    *,
    freeze_only_if_required: Literal[False],
    version_finder: "VersionFinder",
) -> List[NamedRequirement]:
    ...  # pragma: no cover


def freeze(
    requirements: List[NamedRequirement],
    *,
    freeze_only_if_required: bool,
    whitelist: Optional[Whitelist] = None,
    version_finder: "VersionFinder",
) -> List[NamedRequirement]:
    frozen_requirements: List[NamedRequirement] = []

    for requirement in requirements:
        latest_version = _get_latest_version(
            whitelist=whitelist,
            requirement=requirement,
            version_finder=version_finder,
            freeze_only_if_required=freeze_only_if_required,
        )

        if latest_version is not None:
            requirement = NamedRequirement(
                name=requirement.name,
                extras=requirement.extras,
                specs=[f"=={latest_version}"],
                language=requirement.language,
            )

        frozen_requirements.append(requirement)

    return frozen_requirements


def _get_latest_version(
    *,
    whitelist: Optional[Whitelist] = None,
    requirement: NamedRequirement,
    version_finder: "VersionFinder",
    freeze_only_if_required: bool = True,
) -> Optional[str]:
    # don't overwrite existing specs
    if requirement.specs:
        return None

    if freeze_only_if_required:
        assert whitelist is not None

        library = whitelist.find_library(
            language=requirement.language,
            name=requirement.name,
        )

        # not whitelisted?
        if library is None:
            return None

        # not required to be frozen
        if not library.freeze:
            return None

    return version_finder.find_latest(
        name=requirement.name,
    )


class VersionFinder(ABC):

    @abstractmethod
    def find_latest(
        self,
        *,
        language: RequirementLanguage = RequirementLanguage.PYTHON,
        name: str
    ) -> Optional[str]:
        ...  # pragma: no cover


class CrunchHubVersionFinder(VersionFinder):

    def __init__(
        self,
        *,
        api_base_url: str = "https://api.hub.crunchdao.com/",
    ):
        super().__init__()

        self._api_base_url = api_base_url

    def find_latest(
        self,
        *,
        language: RequirementLanguage = RequirementLanguage.PYTHON,
        name: str
    ) -> Optional[str]:
        response = requests.get(
            f"{self._api_base_url}/v1/libraries/{language.name}/{name}/latest",
            params={
                "language": language.value,
                "name": name
            },
        )

        if response.status_code != 200:
            return None

        body = response.json()
        return body.get("version")


class LocalSitePackageVersionFinder(VersionFinder):

    def find_latest(
        self,
        *,
        language: RequirementLanguage = RequirementLanguage.PYTHON,
        name: str
    ) -> Optional[str]:
        if language != RequirementLanguage.PYTHON:
            return None

        from importlib.metadata import PackageNotFoundError, distribution  # late import

        try:
            return distribution(name).version
        except PackageNotFoundError:
            return None



from crunch_convert._model import RequirementLanguage as RequirementLanguage
from crunch_convert.notebook._model import EmbeddedFile as EmbeddedFile
from crunch_convert.notebook._model import ImportedRequirement as ImportedRequirement
from crunch_convert.notebook._notebook import BadCellHandling as BadCellHandling
from crunch_convert.notebook._notebook import ConverterError as ConverterError
from crunch_convert.notebook._notebook import FlattenNotebook as FlattenNotebook
from crunch_convert.notebook._notebook import InconsistantLibraryVersionError as InconsistantLibraryVersionError
from crunch_convert.notebook._notebook import NotebookCellParseError as NotebookCellParseError
from crunch_convert.notebook._notebook import RequirementVersionParseError as RequirementVersionParseError
from crunch_convert.notebook._notebook import extract_from_cells as extract_from_cells
from crunch_convert.notebook._notebook import extract_from_file as extract_from_file

# alias for compatibility with previous versions
EmbedFile = EmbeddedFile
ImportedRequirementLanguage = RequirementLanguage



import json
import os
from typing import Optional, TextIO, cast

import click

from crunch_convert.__version__ import __version__
from crunch_convert._utils import MockedWriteTextIO
from crunch_convert.notebook import ConverterError, InconsistantLibraryVersionError, NotebookCellParseError, extract_from_file
from crunch_convert.notebook._utils import print_indented
from crunch_convert.requirements_txt import CrunchHubWhitelist, LocalSitePackageVersionFinder, format_files_from_imported, format_files_from_named, freeze, parse_from_file


@click.group()
@click.version_option(__version__, package_name="__version__.__title__")
def cli():
    pass  # pragma: no cover


@cli.command(help="Convert a notebook to a python script.")
@click.option("--override", is_flag=True, help="Force overwrite of the python file.")
@click.option("--write-requirements", is_flag=True, help="Write the requirements.txt files.")
@click.option("--write-embedded-files", is_flag=True, help="Write the embedded files.")
@click.option("--no-freeze", is_flag=True, help="Don't freeze the requirements in requirements.txt with locally installed versions.")
@click.argument("notebook-file-path", required=True)
@click.argument("python-file-path", default="main.py")
def notebook(
    override: bool,
    write_requirements: bool,
    write_embedded_files: bool,
    no_freeze: bool,
    notebook_file_path: str,
    python_file_path: str,
):
    print("convert file:", notebook_file_path)

    try:
        flatten = extract_from_file(
            notebook_file_path,
            print=print,
            validate=True,
        )
    except IOError as error:
        print(f"{notebook_file_path}: cannot read notebook file: {error}")
        raise click.Abort()
    except json.JSONDecodeError as error:
        print(f"{notebook_file_path}: cannot parse notebook file: {error}")
        raise click.Abort()
    except ConverterError as error:
        print(f"{notebook_file_path}: convert failed: {error}")

        if isinstance(error, NotebookCellParseError):
            print(f"  cell: {error.cell_id} ({error.cell_index})")
            print(f"  source:")
            print_indented(error.cell_source)
            print(f"  parser error:")
            print_indented(error.parser_error or "None")

        elif isinstance(error, InconsistantLibraryVersionError):
            print(f"  package name: {error.package_name}")
            print(f"  first version: {error.old}")
            print(f"  other version: {error.new}")

        raise click.Abort()

    with _open_with_consent(override, python_file_path) as fd:
        fd.write(flatten.source_code)

    if write_requirements:
        whitelist = CrunchHubWhitelist()
        version_finder = LocalSitePackageVersionFinder()

        requirements_files = format_files_from_imported(
            flatten.requirements,
            header="extracted from a notebook",
            whitelist=whitelist,
        )

        for requirement_language, content in requirements_files.items():
            real_content: str = content
            original_content: Optional[str] = None

            if not no_freeze:
                requirements = parse_from_file(
                    language=requirement_language,
                    file_content=content
                )

                for requirement in requirements:
                    print("freeze requirements:", requirement.name, requirement.extras, requirement.specs)

                frozen_requirements = freeze(
                    requirements=requirements,
                    freeze_only_if_required=False,
                    version_finder=version_finder,
                )

                if requirements != frozen_requirements:
                    frozen_content = format_files_from_named(
                        frozen_requirements,
                        header="frozen from local site-packages",
                        whitelist=whitelist,
                    )

                    real_content = frozen_content[requirement_language]
                    original_content = content

            with _open_with_consent(override, requirement_language.txt_file_name) as fd:
                fd.write(real_content)

            if original_content is not None:
                with _open_with_consent(override, requirement_language.original_txt_file_name) as fd:
                    fd.write(original_content)

    if write_embedded_files:
        for embedded_file in flatten.embedded_files:
            with _open_with_consent(override, embedded_file.normalized_path) as fd:
                fd.write(flatten.source_code)


def _open_with_consent(override: bool, file_path: str) -> TextIO:
    if not override and os.path.exists(file_path):
        override = click.confirm(
            f"file {file_path} already exists, override?",
            default=True,
        )

        if not override:
            return cast(TextIO, MockedWriteTextIO())

    return open(file_path, "w")



from crunch_convert._utils import MockedWriteTextIO


def test_context_manager():
    with MockedWriteTextIO() as mocked_io:
        assert mocked_io is not None
        mocked_io.write(b"hello")  # type: ignore
